import typing
from functools import cache

import typing_extensions

T = typing.TypeVar('T')
P = typing_extensions.ParamSpec('P', bound=typing.Hashable)  # type: ignore


def memoize(func: typing.Callable[P, T]) -> typing.Callable[P, T]:
    return typing.cast(typing.Callable[P, T], cache(func))


@memoize
def as_callable(typ: type[T]) -> typing.Callable[[str], T]:
    """simple wrapper to types to make mypy happy
    use as cast=as_callable(str)"""

    def converter(val: str) -> T:
        return typ(val)  # type: ignore

    return converter



import pathlib
import typing
from dataclasses import dataclass

from config import _helpers
from config.exceptions import MissingName

from .config import MISSING, AnyCallable, Config, T
from .enums import Env

StrOrPath = typing.Union[str, pathlib.Path]
# Validator receives an EnvConfig instance and the env_name currently checking
Validator = typing.Callable[['Config'], typing.Callable[[str], bool]]

# Using tuples with integers to prevent bugs with sorting


class OptionalConfig(Config):
    def __init__(
        self,
        required_if: Validator,
        env_file: typing.Optional[StrOrPath],
        mapping: typing.Optional[typing.Mapping[str, str]] = None,
        ignore_default: bool = True,
    ) -> None:
        super().__init__(env_file)
        self._mapping = mapping or self._mapping
        self._ignore_default = ignore_default
        self._required_if = required_if(Config(mapping=self._mapping))

    @_helpers.memoize
    def validate(self, name: str):
        return self._required_if(name)

    def _get_value(
        self,
        name: str,
        default: typing.Any,
        ignore_default: typing.Optional[bool] = None,
    ) -> str:
        ignore_default = (
            self._ignore_default if ignore_default is None else ignore_default
        )
        value: typing.Union[str, object] = self._mapping.get(name, default)
        if self.validate(name) and value is MISSING:
            value = self._file_vals.get(name, MISSING)
        if value is MISSING and ignore_default:
            raise MissingName(name)
        return typing.cast(str, value)

    def get(
        self,
        name: str,
        cast: typing.Optional[AnyCallable] = None,
        default: typing.Any = MISSING,
        ignore_default: typing.Optional[bool] = None,
    ) -> typing.Any:
        value = self._get_value(name, default, ignore_default)
        return value if cast is None else self._cast(name, value, cast)

    @typing.overload
    def __call__(
        self,
        name: str,
        cast: typing.Callable[[str], T],
        default: typing.Any = MISSING,
        ignore_default: typing.Optional[bool] = None,
    ) -> T:
        ...

    @typing.overload
    def __call__(
        self,
        name: str,
        cast: None = None,
        default: typing.Any = MISSING,
        ignore_default: typing.Optional[bool] = None,
    ) -> str:
        ...

    def __call__(
        self,
        name: str,
        cast: typing.Optional[AnyCallable] = None,
        default: typing.Any = MISSING,
        ignore_default: typing.Optional[bool] = None,
    ) -> typing.Any:
        return self.get(name, cast, default, ignore_default)


ENV_RELEVANCY = {Env.TEST: 0, Env.LOCAL: 1, Env.DEV: 2, Env.PRD: 3}


@dataclass(frozen=True)
class ByRelevancy:

    max_relevancy: Env = Env.LOCAL
    env_name: str = 'ENV'

    if typing.TYPE_CHECKING:

        @property
        def env(self) -> Env:
            ...

    def required_if(self, config: Config):
        env = config(self.env_name, Env)
        object.__setattr__(self, 'env', env)
        expected = ENV_RELEVANCY[self.max_relevancy]

        def validator(_: str):
            return ENV_RELEVANCY[env] <= expected

        return validator


class EnvConfig(OptionalConfig):
    def __init__(
        self,
        by_relevancy: ByRelevancy = ByRelevancy(),
        env_file: StrOrPath = '.env',
        mapping: typing.Optional[typing.Mapping[str, str]] = None,
        ignore_default: bool = True,
    ) -> None:
        self._by_relevancy = by_relevancy
        super().__init__(
            by_relevancy.required_if, env_file, mapping, ignore_default
        )

    @property
    def env(self) -> Env:
        return self._by_relevancy.env



import typing
from collections.abc import Callable
from shlex import shlex

from ._helpers import as_callable

T = typing.TypeVar('T')


@typing.overload
def comma_separated(
    cast: Callable[[str], str] = as_callable(str)
) -> typing.Callable[[str], tuple[str, ...]]:
    ...


@typing.overload
def comma_separated(
    cast: Callable[[str], T]
) -> typing.Callable[[str], tuple[T, ...]]:
    ...


def comma_separated(
    cast: Callable[[str], typing.Union[T, str]] = as_callable(str)
) -> typing.Callable[[str], tuple[typing.Union[T, str], ...]]:
    def _wrapped(val: str) -> tuple[typing.Union[T, str], ...]:
        lex = shlex(val, posix=True)
        lex.whitespace = ','
        lex.whitespace_split = True
        return tuple(cast(item.strip()) for item in lex)

    return _wrapped


_boolean_vals = {
    'true': True,
    'false': False,
    '0': False,
    '1': True,
}


def boolean(val: typing.Optional[str]) -> bool:
    if not val:
        return False
    try:
        return _boolean_vals[val.lower()]
    except KeyError as err:
        raise ValueError(
            f'Received invalid bool: {val}, try: {",".join(_boolean_vals)}'
        ) from err



from . import enums, helpers
from ._helpers import as_callable
from .config import MISSING, CachedConfig, CIConfig, Config
from .enums import Env
from .envconfig import EnvConfig
from .exceptions import AlreadySet, InvalidCast, MissingName
from .mapping import EnvMapping, LowerEnvMapping

__all__ = [
    'as_callable',
    'Config',
    'CachedConfig',
    'CIConfig',
    'MISSING',
    'EnvMapping',
    'LowerEnvMapping',
    'Env',
    'MissingName',
    'InvalidCast',
    'EnvConfig',
    'AlreadySet',
    'enums',
    'helpers',
]


__version__ = '1.8.0'



from enum import Enum


class Env(Enum):
    LOCAL = 'local'
    DEV = 'dev'
    TEST = 'test'
    PRD = 'prd'

    @classmethod
    def iter(cls):
        for item in cls:
            yield item.value
